@@ -30,6 +30,7 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 #include "ui/window.h"
 
 #include <SDL_timer.h>
+#include <SDL_syswm.h>
 
 #import <AppKit/AppKit.h>
 
@@ -51,6 +52,16 @@ static iInt2 macVer_(void) {
     return init_I2(10, 10);
 }
 
+static NSWindow *nsWindow_(SDL_Window *window) {
+    SDL_SysWMinfo wm;
+    SDL_VERSION(&wm.version);
+    if (SDL_GetWindowWMInfo(window, &wm)) {
+        return wm.info.cocoa.window;
+    }
+    iAssert(false);
+    return nil;
+}
+
 static NSString *currentSystemAppearance_(void) {
     /* This API does not exist on 10.13. */
     if ([NSApp respondsToSelector:@selector(effectiveAppearance)]) {
@@ -66,6 +77,14 @@ iBool shouldDefaultToMetalRenderer_MacOS(void) {
     return ver.x > 10 || ver.y > 13;*/
 }
 
+static void ignoreImmediateKeyDownEvents_(void) {
+    /* SDL ignores menu key equivalents so the keydown events will be posted regardless.
+       However, we shouldn't double-activate menu items when a shortcut key is used in our
+       widgets. Quite a kludge: take advantage of Window's focus-acquisition threshold to
+       ignore the immediately following key down events. */
+    get_Window()->focusGainedAt = SDL_GetTicks();
+}
+
 /*----------------------------------------------------------------------------------------------*/
 
 @interface CommandButton : NSCustomTouchBarItem {
@@ -135,11 +154,60 @@ iBool shouldDefaultToMetalRenderer_MacOS(void) {
 
 /*----------------------------------------------------------------------------------------------*/
 
+@interface MenuCommands : NSObject {
+    NSMutableDictionary<NSString *, NSString *> *commands;
+    iWidget *source;
+}
+@end
+
+@implementation MenuCommands
+
+- (id)init {
+    commands = [[NSMutableDictionary<NSString *, NSString *> alloc] init];
+    source = NULL;
+    return self;
+}
+
+- (void)setCommand:(NSString *)command forMenuItem:(NSMenuItem *)menuItem {
+    [commands setObject:command forKey:[menuItem title]];
+}
+
+- (void)setSource:(iWidget *)widget {
+    source = widget;
+}
+
+- (void)clear {
+    [commands removeAllObjects];
+}
+
+- (NSString *)commandForMenuItem:(NSMenuItem *)menuItem {
+    return [commands objectForKey:[menuItem title]];
+}
+
+- (void)postMenuItemCommand:(id)sender {
+    NSString *command = [commands objectForKey:[(NSMenuItem *)sender title]];
+    if (command) {
+        const char *cstr = [command cStringUsingEncoding:NSUTF8StringEncoding];
+        if (source) {
+            postCommand_Widget(source, "%s", cstr);
+        }
+        else {
+            postCommand_Root(NULL, cstr);
+        }
+        ignoreImmediateKeyDownEvents_();
+    }
+}
+
+@end
+
+/*----------------------------------------------------------------------------------------------*/
+
 @interface MyDelegate : NSResponder<NSApplicationDelegate, NSTouchBarDelegate> {
     enum iTouchBarVariant touchBarVariant;
     NSString *currentAppearanceName;
     NSObject<NSApplicationDelegate> *sdlDelegate;
-    NSMutableDictionary<NSString *, NSString*> *menuCommands;
+    //NSMutableDictionary<NSString *, NSString*> *menuCommands;
+    MenuCommands *menuCommands;
 }
 - (id)initWithSDLDelegate:(NSObject<NSApplicationDelegate> *)sdl;
 - (NSTouchBar *)makeTouchBar;
@@ -154,7 +222,7 @@ iBool shouldDefaultToMetalRenderer_MacOS(void) {
 - (id)initWithSDLDelegate:(NSObject<NSApplicationDelegate> *)sdl {
     [super init];
     currentAppearanceName = nil;
-    menuCommands = [[NSMutableDictionary<NSString *, NSString *> alloc] init];
+    menuCommands = [[MenuCommands alloc] init];
     touchBarVariant = default_TouchBarVariant;
     sdlDelegate = sdl;
     return self;
@@ -171,6 +239,14 @@ iBool shouldDefaultToMetalRenderer_MacOS(void) {
     self.touchBar = nil;
 }
 
+- (MenuCommands *)menuCommands {
+    return menuCommands;
+}
+
+- (void)postMenuItemCommand:(id)sender {
+    [menuCommands postMenuItemCommand:sender];
+}
+
 static void appearanceChanged_MacOS_(NSString *name) {
     const iBool isDark = [name containsString:@"Dark"];
     const iBool isHighContrast = [name containsString:@"HighContrast"];
@@ -187,10 +263,6 @@ static void appearanceChanged_MacOS_(NSString *name) {
     }
 }
 
-- (void)setCommand:(NSString *)command forMenuItem:(NSMenuItem *)menuItem {
-    [menuCommands setObject:command forKey:[menuItem title]];
-}
-
 - (BOOL)application:(NSApplication *)app openFile:(NSString *)filename {
     return [sdlDelegate application:app openFile:filename];
 }
@@ -247,31 +319,11 @@ static void appearanceChanged_MacOS_(NSString *name) {
     ignoreImmediateKeyDownEvents_();
 }
 
-static void ignoreImmediateKeyDownEvents_(void) {
-    /* SDL ignores menu key equivalents so the keydown events will be posted regardless.
-       However, we shouldn't double-activate menu items when a shortcut key is used in our
-       widgets. Quite a kludge: take advantage of Window's focus-acquisition threshold to
-       ignore the immediately following key down events. */
-    get_Window()->focusGainedAt = SDL_GetTicks();
-}
-
 - (void)closeTab {
     postCommand_App("tabs.close");
     ignoreImmediateKeyDownEvents_();
 }
 
-- (NSString *)commandForItem:(NSMenuItem *)menuItem {
-    return [menuCommands objectForKey:[menuItem title]];
-}
-
-- (void)postMenuItemCommand:(id)sender {
-    NSString *command = [menuCommands objectForKey:[(NSMenuItem *)sender title]];
-    if (command) {
-        postCommand_App([command cStringUsingEncoding:NSUTF8StringEncoding]);
-        ignoreImmediateKeyDownEvents_();
-    }
-}
-
 - (void)sidebarModePressed:(id)sender {
     NSSegmentedControl *seg = sender;
     postCommandf_App("sidebar.mode arg:%d toggle:1", (int) [seg selectedSegment]);
@@ -370,6 +422,11 @@ void setupApplication_MacOS(void) {
     windowCloseItem.action = @selector(closeTab);
 }
 
+void hideTitleBar_MacOS(iWindow *window) {
+    NSWindow *w = nsWindow_(window->win);
+    w.styleMask = 0; /* borderless */
+}
+
 void enableMenu_MacOS(const char *menuLabel, iBool enable) {
     menuLabel = translateCStr_Lang(menuLabel);
     NSApplication *app = [NSApplication sharedApplication];
@@ -377,7 +434,6 @@ void enableMenu_MacOS(const char *menuLabel, iBool enable) {
     NSString *label = [NSString stringWithUTF8String:menuLabel];
     NSMenuItem *menuItem = [appMenu itemAtIndex:[appMenu indexOfItemWithTitle:label]];
     [menuItem setEnabled:enable];
-    [label release];
 }
 
 void enableMenuItem_MacOS(const char *menuItemCommand, iBool enable) {
@@ -388,7 +444,7 @@ void enableMenuItem_MacOS(const char *menuItemCommand, iBool enable) {
         NSMenu *menu = mainMenuItem.submenu;
         if (menu) {
             for (NSMenuItem *menuItem in menu.itemArray) {
-                NSString *command = [myDel commandForItem:menuItem];
+                NSString *command = [[myDel menuCommands] commandForMenuItem:menuItem];
                 if (command) {
                     if (!iCmpStr([command cStringUsingEncoding:NSUTF8StringEncoding],
                                  menuItemCommand)) {
@@ -468,6 +524,74 @@ void removeMenu_MacOS(int atIndex) {
     [appMenu removeItemAtIndex:atIndex];
 }
 
+enum iColorId removeColorEscapes_String(iString *d) {
+    enum iColorId color = none_ColorId;
+    for (;;) {
+        const char *esc = strchr(cstr_String(d), '\v');
+        if (esc) {
+            const char *endp;
+            color = parseEscape_Color(esc, &endp);
+            remove_Block(&d->chars, esc - cstr_String(d), endp - esc);
+        }
+        else break;
+    }
+    return color;
+}
+
+// returns the selected item, if any
+static NSMenuItem *makeMenuItems_(NSMenu *menu, MenuCommands *commands, const iMenuItem *items, size_t n) {
+    NSMenuItem *selectedItem = nil;
+    for (size_t i = 0; i < n && items[i].label; ++i) {
+        const char *label = translateCStr_Lang(items[i].label);
+        if (equal_CStr(label, "---")) {
+            [menu addItem:[NSMenuItem separatorItem]];
+        }
+        else {
+            const iBool hasCommand = (items[i].command && items[i].command[0]);
+            iBool isChecked = iFalse;
+            iBool isDisabled = iFalse;
+            if (startsWith_CStr(label, "###")) {
+                isChecked = iTrue;
+                label += 3;
+            }
+            else if (startsWith_CStr(label, "///")) {
+                isDisabled = iTrue;
+                label += 3;
+            }
+            iString itemTitle;
+            initCStr_String(&itemTitle, label);
+            removeIconPrefix_String(&itemTitle);
+            if (removeColorEscapes_String(&itemTitle) == uiTextCaution_ColorId) {
+//                prependCStr_String(&itemTitle, "\u26a0\ufe0f ");
+            }
+            NSMenuItem *item = [menu addItemWithTitle:[NSString stringWithUTF8String:cstr_String(&itemTitle)]
+                                               action:(hasCommand ? @selector(postMenuItemCommand:) : nil)
+                                        keyEquivalent:@""];
+            deinit_String(&itemTitle);
+            [item setTarget:commands];
+            if (isChecked) {
+                [item setState:NSControlStateValueOn];
+                selectedItem = item;
+            }
+            [item setEnabled:!isDisabled];
+            int key   = items[i].key;
+            int kmods = items[i].kmods;
+            if (hasCommand) {
+                [commands setCommand:[NSString stringWithUTF8String:items[i].command]
+                         forMenuItem:item];
+                /* Bindings may have a different key. */
+                const iBinding *bind = findCommand_Keys(items[i].command);
+                if (bind && bind->id < builtIn_BindingId) {
+                    key   = bind->key;
+                    kmods = bind->mods;
+                }
+            }
+            setShortcut_NSMenuItem_(item, key, kmods);
+        }
+    }
+    return selectedItem;
+}
+
 void insertMenuItems_MacOS(const char *menuLabel, int atIndex, const iMenuItem *items, size_t count) {
     NSApplication *app = [NSApplication sharedApplication];
     MyDelegate *myDel = (MyDelegate *) app.delegate;
@@ -479,34 +603,7 @@ void insertMenuItems_MacOS(const char *menuLabel, int atIndex, const iMenuItem *
                                                 atIndex:atIndex];
     NSMenu *menu = [[NSMenu alloc] initWithTitle:[NSString stringWithUTF8String:menuLabel]];
     [menu setAutoenablesItems:NO];
-    for (size_t i = 0; i < count; ++i) {
-        const char *label = translateCStr_Lang(items[i].label);
-        if (label[0] == '\v') {
-            /* Skip the formatting escape. */
-            label += 2;
-        }
-        if (equal_CStr(label, "---")) {
-            [menu addItem:[NSMenuItem separatorItem]];
-        }
-        else {
-            const iBool hasCommand = (items[i].command && items[i].command[0]);
-            NSMenuItem *item = [menu addItemWithTitle:[NSString stringWithUTF8String:label]
-                                               action:(hasCommand ? @selector(postMenuItemCommand:) : nil)
-                                        keyEquivalent:@""];
-            int key   = items[i].key;
-            int kmods = items[i].kmods;
-            if (hasCommand) {
-                [myDel setCommand:[NSString stringWithUTF8String:items[i].command] forMenuItem:item];
-                /* Bindings may have a different key. */
-                const iBinding *bind = findCommand_Keys(items[i].command);
-                if (bind && bind->id < builtIn_BindingId) {
-                    key   = bind->key;
-                    kmods = bind->mods;
-                }
-            }
-            setShortcut_NSMenuItem_(item, key, kmods);
-        }
-    }
+    makeMenuItems_(menu, [myDel menuCommands], items, count);
     [mainItem setSubmenu:menu];
     [menu release];
 }
@@ -527,7 +624,7 @@ void handleCommand_MacOS(const char *cmd) {
             if (menu) {
                 int itemIndex = 0;
                 for (NSMenuItem *menuItem in menu.itemArray) {
-                    NSString *command = [myDel commandForItem:menuItem];
+                    NSString *command = [[myDel menuCommands] commandForMenuItem:menuItem];
                     if (!command && mainIndex == 6 && itemIndex == 0) {
                         /* Window > Close */
                         command = @"tabs.close";
@@ -553,3 +650,49 @@ void handleCommand_MacOS(const char *cmd) {
 void log_MacOS(const char *msg) {
     NSLog(@"%s", msg);
 }
+
+void showPopupMenu_MacOS(iWidget *source, iInt2 windowCoord, const iMenuItem *items, size_t n) {
+    NSMenu *      menu         = [[NSMenu alloc] init];
+    MenuCommands *menuCommands = [[MenuCommands alloc] init];
+    iWindow *     window       = as_Window(mainWindow_App());
+    NSWindow *    nsWindow     = nsWindow_(window->win);
+    /* View coordinates are flipped. */
+    iBool isCentered = iFalse;
+    if (isEqual_I2(windowCoord, zero_I2())) {
+        windowCoord = divi_I2(window->size, 2);
+        isCentered = iTrue;
+    }
+    windowCoord.y = window->size.y - windowCoord.y;
+    windowCoord = divf_I2(windowCoord, window->pixelRatio);
+    NSPoint screenPoint = [nsWindow convertPointToScreen:(CGPoint){ windowCoord.x, windowCoord.y }];
+    NSMenuItem *selectedItem = makeMenuItems_(menu, menuCommands, items, n);
+    [menuCommands setSource:source];
+    if (isCentered) {
+        NSSize menuSize = [menu size];
+        screenPoint.x -= menuSize.width / 2;
+        screenPoint.y += menuSize.height / 2;
+    }
+    [menu setAutoenablesItems:NO];
+    [menu popUpMenuPositioningItem:selectedItem atLocation:screenPoint inView:nil];
+    [menu release];
+    [menuCommands release];
+    /* The right mouse button has now been released so let SDL know about it. The button up event
+       was consumed by the popup menu so it got never passed to SDL. Same goes for possible
+       keyboard modifiers that were held down when the menu appeared. */
+    SEL sel = NSSelectorFromString(@"syncMouseButtonAndKeyboardModifierState"); /* custom method */
+    if ([[nsWindow delegate] respondsToSelector:sel]) {
+        NSInvocation *call = [NSInvocation invocationWithMethodSignature:
+                              [NSMethodSignature signatureWithObjCTypes:"v@:"]];
+        [call setSelector:sel];
+        [call invokeWithTarget:[nsWindow delegate]];
+    }
+}
+
+iColor systemAccent_Color(void) {
+    NSColor *accent = [[NSColor controlAccentColor] colorUsingColorSpace:
+                                                        [NSColorSpace deviceRGBColorSpace]];
+    return (iColor){ iClamp([accent redComponent]   * 255, 0, 255),
+                     iClamp([accent greenComponent] * 255, 0, 255),
+                     iClamp([accent blueComponent]  * 255, 0, 255),
+                     255 };
+}
