@@ -24,11 +24,13 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 #include "root.h"
 #include "app.h"
 
+#include <the_Foundation/file.h>
+#include <the_Foundation/path.h>
 #include <the_Foundation/string.h>
 
 static const iColor transparent_;
 
-static const iColor darkPalette_[] = {
+static iColor darkPalette_[] = {
     { 0,   0,   0,   255 },
     { 40,  40,  40,  255 },
     { 80,  80,  80,  255 },
@@ -47,7 +49,7 @@ static const iColor darkPalette_[] = {
     { 0,   200, 0,   255 },
 };
 
-static const iColor lightPalette_[] = {
+static iColor lightPalette_[] = {
     { 0,   0,   0,   255 },
     { 75,  75,  75,  255 },
     { 150, 150, 150, 255 },
@@ -79,10 +81,17 @@ iLocalDef void copy_(enum iColorId dst, enum iColorId src) {
 void setThemePalette_Color(enum iColorTheme theme) {
     const iPrefs *prefs = prefs_App();
     memcpy(uiPalette_, isDark_ColorTheme(theme) ? darkPalette_ : lightPalette_, sizeof(darkPalette_));
+    if (prefs->accent == system_ColorAccent) {
+        iColor systemColor = systemAccent_Color();
+        memcpy(&uiPalette_[cyan_ColorId], &systemColor, sizeof(iColor));
+        memcpy(&uiPalette_[orange_ColorId], &systemColor, sizeof(iColor));
+    }
     const int accentHi    = (prefs->accent == cyan_ColorAccent ? cyan_ColorId : orange_ColorId);
     const int accentLo    = (prefs->accent == cyan_ColorAccent ? teal_ColorId : brown_ColorId);
     const int altAccentHi = (prefs->accent == cyan_ColorAccent ? orange_ColorId : cyan_ColorId);
     const int altAccentLo = (prefs->accent == cyan_ColorAccent ? brown_ColorId : teal_ColorId);
+    const iColor accentMid    = mix_Color(get_Color(accentHi), get_Color(accentLo), 0.5f);
+    const iColor altAccentMid = mix_Color(get_Color(altAccentHi), get_Color(altAccentLo), 0.5f);
     switch (theme) {
         case pureBlack_ColorTheme: {
             copy_(uiBackground_ColorId, black_ColorId);
@@ -118,10 +127,11 @@ void setThemePalette_Color(enum iColorTheme theme) {
             copy_(uiInputBackground_ColorId, black_ColorId);
             copy_(uiInputBackgroundFocused_ColorId, black_ColorId);
             copy_(uiInputText_ColorId, gray75_ColorId);
-            copy_(uiInputTextFocused_ColorId, white_ColorId);
+            set_Color(uiInputTextFocused_ColorId, mix_Color(get_Color(white_ColorId),
+                                                            get_Color(accentHi), 0.15f));
             copy_(uiInputFrame_ColorId, gray25_ColorId);
             copy_(uiInputFrameHover_ColorId, accentHi);
-            copy_(uiInputFrameFocused_ColorId, altAccentHi);
+            copy_(uiInputFrameFocused_ColorId, accentLo);
             copy_(uiInputCursor_ColorId, altAccentHi);
             copy_(uiInputCursorText_ColorId, black_ColorId);
             copy_(uiHeading_ColorId, accentHi);
@@ -129,8 +139,8 @@ void setThemePalette_Color(enum iColorTheme theme) {
             copy_(uiIcon_ColorId, accentHi);
             copy_(uiIconHover_ColorId, accentHi);
             copy_(uiSeparator_ColorId, gray25_ColorId);
-            copy_(uiMarked_ColorId, altAccentLo);
-            copy_(uiMatching_ColorId, accentLo);
+            copy_(uiMarked_ColorId, accentLo);
+            copy_(uiMatching_ColorId, altAccentLo);
             break;
         }
         default:
@@ -169,10 +179,12 @@ void setThemePalette_Color(enum iColorTheme theme) {
                       mix_Color(get_Color(black_ColorId), get_Color(gray25_ColorId), 0.7f));
             copy_(uiInputBackgroundFocused_ColorId, black_ColorId);
             copy_(uiInputText_ColorId, gray75_ColorId);
-            copy_(uiInputTextFocused_ColorId, white_ColorId);
+            //copy_(uiInputTextFocused_ColorId, white_ColorId);
+            set_Color(uiInputTextFocused_ColorId, mix_Color(get_Color(white_ColorId),
+                                                            get_Color(accentHi), 0.15f));
             copy_(uiInputFrame_ColorId, uiInputBackground_ColorId);
             copy_(uiInputFrameHover_ColorId, accentHi);
-            copy_(uiInputFrameFocused_ColorId, altAccentHi);
+            copy_(uiInputFrameFocused_ColorId, accentLo);
             copy_(uiInputCursor_ColorId, altAccentHi);
             copy_(uiInputCursorText_ColorId, black_ColorId);
             copy_(uiHeading_ColorId, accentHi);
@@ -180,8 +192,8 @@ void setThemePalette_Color(enum iColorTheme theme) {
             copy_(uiIcon_ColorId, accentHi);
             copy_(uiIconHover_ColorId, accentHi);
             copy_(uiSeparator_ColorId, black_ColorId);
-            copy_(uiMarked_ColorId, altAccentLo);
-            copy_(uiMatching_ColorId, accentLo);
+            copy_(uiMarked_ColorId, accentLo);
+            copy_(uiMatching_ColorId, altAccentLo);
             break;
         }
         case light_ColorTheme:
@@ -222,7 +234,7 @@ void setThemePalette_Color(enum iColorTheme theme) {
             set_Color(uiInputFrame_ColorId,
                       mix_Color(get_Color(gray50_ColorId), get_Color(gray75_ColorId), 0.5f));
             copy_(uiInputFrameHover_ColorId, accentLo);
-            copy_(uiInputFrameFocused_ColorId, altAccentLo);
+            copy_(uiInputFrameFocused_ColorId, accentLo);
             copy_(uiInputCursor_ColorId, altAccentLo);
             copy_(uiInputCursorText_ColorId, white_ColorId);
             copy_(uiHeading_ColorId, accentLo);
@@ -231,8 +243,8 @@ void setThemePalette_Color(enum iColorTheme theme) {
             copy_(uiIconHover_ColorId, accentLo);
             set_Color(uiSeparator_ColorId,
                       mix_Color(get_Color(gray50_ColorId), get_Color(gray75_ColorId), 0.5f));
-            copy_(uiMarked_ColorId, altAccentHi);
-            copy_(uiMatching_ColorId, accentHi);
+            copy_(uiMarked_ColorId, accentHi);
+            copy_(uiMatching_ColorId, altAccentHi);
             break;
         case pureWhite_ColorTheme:
             copy_(uiBackground_ColorId, white_ColorId);
@@ -273,7 +285,7 @@ void setThemePalette_Color(enum iColorTheme theme) {
             copy_(uiInputTextFocused_ColorId, black_ColorId);
             copy_(uiInputFrame_ColorId, gray50_ColorId);
             copy_(uiInputFrameHover_ColorId, accentLo);
-            copy_(uiInputFrameFocused_ColorId, altAccentLo);
+            copy_(uiInputFrameFocused_ColorId, accentLo);
             copy_(uiInputCursor_ColorId, altAccentLo);
             copy_(uiInputCursorText_ColorId, white_ColorId);
             copy_(uiHeading_ColorId, accentLo);
@@ -282,8 +294,8 @@ void setThemePalette_Color(enum iColorTheme theme) {
             copy_(uiIconHover_ColorId, accentLo);
             set_Color(uiSeparator_ColorId,
                       mix_Color(get_Color(gray50_ColorId), get_Color(gray75_ColorId), 0.67f));
-            copy_(uiMarked_ColorId, altAccentHi);
-            copy_(uiMatching_ColorId, accentHi);
+            copy_(uiMarked_ColorId, accentHi);
+            copy_(uiMatching_ColorId, altAccentHi);
             break;
     }
     set_Color(uiSubheading_ColorId,
@@ -468,12 +480,30 @@ const char *escape_Color(int color) {
         return esc[color];
     }
     /* TODO: Conflict with format strings! "%" (37) may be used as the color value. */
-    /* Double-\r is used for range extension. */
+    /* Double-\v is used for range extension. */
     if (color + asciiBase_ColorEscape > 127) {
         iAssert(color - asciiExtended_ColorEscape + asciiBase_ColorEscape <= 127);
-        return format_CStr("\r\r%c", color - asciiExtended_ColorEscape + asciiBase_ColorEscape);
+        return format_CStr("\v\v%c", color - asciiExtended_ColorEscape + asciiBase_ColorEscape);
     }
-    return format_CStr("\r%c", color + asciiBase_ColorEscape);
+    return format_CStr("\v%c", color + asciiBase_ColorEscape);
+}
+
+enum iColorId parseEscape_Color(const char *cstr, const char **endp) {
+    enum iColorId color = none_ColorId;
+    if (*cstr == '\v') {
+        cstr++;
+        color = 0;
+        if (*cstr == '\v') {
+            color += asciiExtended_ColorEscape;
+            cstr++;
+        }
+        color += *cstr - asciiBase_ColorEscape;
+        cstr++;
+    }
+    if (endp) {
+        *endp = cstr;
+    }
+    return color;
 }
 
 iHSLColor setSat_HSLColor(iHSLColor d, float sat) {
@@ -804,3 +834,83 @@ iColor ansiForeground_Color(iRangecc escapeSequence, int fallback) {
     }
     return clr;
 }
+
+iBool loadPalette_Color(const char *path) {
+    iBool wasLoaded = iFalse;
+    iFile *f = newCStr_File(concatPath_CStr(path, "palette.txt"));
+    if (open_File(f, text_FileMode | readOnly_FileMode)) {
+        iColor *dstPal = darkPalette_;
+        iRangecc srcLine = iNullRange;
+        const iBlock *src = collect_Block(readAll_File(f));
+        while (nextSplit_Rangecc(range_Block(src), "\n", &srcLine)) {
+            iRangecc line = srcLine;
+            trim_Rangecc(&line);
+            if (isEmpty_Range(&line)) {
+                continue;
+            }
+            if (*line.start == '#') {
+                /* Control directive. */
+                line.start++;
+                trim_Rangecc(&line);
+                if (equalCase_Rangecc(line, "dark")) {
+                    dstPal = darkPalette_;
+                }
+                else if (equalCase_Rangecc(line, "light")) {
+                    dstPal = lightPalette_;
+                }
+                continue;
+            }
+            static const struct {
+                const char *label;
+                int         paletteIndex;
+            } colors_[] = {
+                { "black:", 0 }, { "gray25:", 1 }, { "gray50:", 2 }, { "gray75:", 3 },
+                { "white:", 4 }, { "brown:", 5 },  { "orange:", 6 }, { "teal:", 7 },
+                { "cyan:", 8 },  { "yellow:", 9 }, { "red:", 10 },   { "magenta:", 11 },
+                { "blue:", 12 }, { "green:", 13 },
+            };
+            iForIndices(i, colors_) {
+                if (startsWithCase_Rangecc(line, colors_[i].label)) {
+                    iColor *dst = &dstPal[colors_[i].paletteIndex];
+                    line.start += strlen(colors_[i].label);
+                    trim_Rangecc(&line);
+                    if (!isEmpty_Range(&line)) {
+                        if (*line.start == '#') {
+                            /* Hexadecimal color. */
+                            line.start++;
+                            if (size_Range(&line) == 6) {
+                                iBlock *vals = hexDecode_Rangecc(line);
+                                iAssert(size_Block(vals) == 3);
+                                const uint8_t *rgb = constData_Block(vals);
+                                *dst = (iColor){ rgb[0], rgb[1], rgb[2], 255 };
+                                delete_Block(vals);
+                            }
+                            else {
+                                fprintf(stderr, "[Color] invalid custom color: %s\n",
+                                        cstr_Rangecc(line));
+                            }
+                        }
+                        else {
+                            unsigned int red = 0, green = 0, blue = 0;
+                            sscanf(line.start, "%u %u %u", &red, &green, &blue);
+                            if (red > 255 || green > 255 || blue > 255) {
+                                fprintf(stderr, "[Color] RGB value(s) out of range: %s\n",
+                                        cstr_Rangecc(line));
+                            }
+                            *dst = (iColor){ red, green, blue, 255 };
+                        }
+                    }
+                }
+            }
+        }
+        wasLoaded = iTrue;
+    }
+    iRelease(f);
+    return wasLoaded;
+}
+
+#if !defined (iPlatformAppleDesktop)
+iColor systemAccent_Color(void) {
+    return (iColor){ 255, 255, 255, 255 };
+}
+#endif
