@@ -32,8 +32,10 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 #include <the_Foundation/intset.h>
 
 void init_ListItem(iListItem *d) {
-    d->isSeparator = iFalse;
-    d->isSelected  = iFalse;
+    d->isSeparator  = iFalse;
+    d->isSelected   = iFalse;
+    d->isDraggable  = iFalse;
+    d->isDropTarget = iFalse;
 }
 
 void deinit_ListItem(iListItem *d) {
@@ -54,6 +56,8 @@ struct Impl_ListWidget {
     int itemHeight;
     iPtrArray items;
     size_t hoverItem;
+    size_t dragItem;
+    iInt2 dragOrigin; /* offset from mouse to drag item's top-left corner */
     iClick click;
     iIntSet invalidItems;
     iVisBuf *visBuf;
@@ -95,6 +99,8 @@ void init_ListWidget(iListWidget *d) {
     d->noHoverWhileScrolling = iFalse;
     init_PtrArray(&d->items);
     d->hoverItem = iInvalidPos;
+    d->dragItem = iInvalidPos;
+    d->dragOrigin = zero_I2();
     init_Click(&d->click, d, SDL_BUTTON_LEFT);
     init_IntSet(&d->invalidItems);
     d->visBuf = new_VisBuf();
@@ -248,6 +254,10 @@ const iAnyObject *constItem_ListWidget(const iListWidget *d, size_t index) {
     return NULL;
 }
 
+const iAnyObject *constDragItem_ListWidget(const iListWidget *d) {
+    return constItem_ListWidget(d, d->dragItem);
+}
+
 const iAnyObject *constHoverItem_ListWidget(const iListWidget *d) {
     return constItem_ListWidget(d, d->hoverItem);
 }
@@ -267,7 +277,7 @@ size_t hoverItemIndex_ListWidget(const iListWidget *d) {
     return d->hoverItem;
 }
 
-static void setHoverItem_ListWidget_(iListWidget *d, size_t index) {
+void setHoverItem_ListWidget(iListWidget *d, size_t index) {
     if (index < size_PtrArray(&d->items)) {
         const iListItem *item = at_PtrArray(&d->items, index);
         if (item->isSeparator) {
@@ -283,8 +293,8 @@ static void setHoverItem_ListWidget_(iListWidget *d, size_t index) {
 }
 
 void updateMouseHover_ListWidget(iListWidget *d) {
-    const iInt2 mouse = mouseCoord_Window(get_Window());
-    setHoverItem_ListWidget_(d, itemIndex_ListWidget(d, mouse));
+    const iInt2 mouse = mouseCoord_Window(get_Window(), 0);
+    setHoverItem_ListWidget(d, itemIndex_ListWidget(d, mouse));
 }
 
 void sort_ListWidget(iListWidget *d, int (*cmp)(const iListItem **item1, const iListItem **item2)) {
@@ -308,7 +318,51 @@ static void updateHover_ListWidget_(iListWidget *d, const iInt2 mouse) {
         contains_Widget(constAs_Widget(d), mouse)) {
         hover = itemIndex_ListWidget(d, mouse);
     }
-    setHoverItem_ListWidget_(d, hover);
+    setHoverItem_ListWidget(d, hover);
+}
+
+static size_t resolveDragDestination_ListWidget_(const iListWidget *d, iInt2 dstPos, iBool *isOnto) {
+    size_t           index = itemIndex_ListWidget(d, dstPos);
+    const iListItem *item  = constItem_ListWidget(d, index);
+    if (!item) {
+        index = (dstPos.y < mid_Rect(bounds_Widget(constAs_Widget(d))).y ? 0 : (numItems_ListWidget(d) - 1));
+        item = constItem_ListWidget(d, index);
+    }
+    const iRect   rect = itemRect_ListWidget(d, index);
+    const iRangei span = ySpan_Rect(rect);
+    if (item->isDropTarget) {
+        const int pad = size_Range(&span) / 4;
+        if (dstPos.y >= span.start + pad && dstPos.y < span.end - pad) {
+            *isOnto = iTrue;
+            return index;
+        }
+    }
+    if (dstPos.y - span.start > span.end - dstPos.y) {
+        index++;
+    }
+    index = iMin(index, numItems_ListWidget(d));
+    *isOnto = iFalse;
+    return index;
+}
+
+static iBool endDrag_ListWidget_(iListWidget *d, iInt2 endPos) {
+    if (d->dragItem == iInvalidPos) {
+        return iFalse;
+    }
+    stop_Anim(&d->scrollY.pos);
+    iBool isOnto;
+    const size_t index = resolveDragDestination_ListWidget_(d, endPos, &isOnto);
+    if (index != d->dragItem) {
+        if (isOnto) {
+            postCommand_Widget(d, "list.dragged arg:%zu onto:%zu", d->dragItem, index);
+        }
+        else {
+            postCommand_Widget(d, "list.dragged arg:%zu before:%zu", d->dragItem, index);
+        }
+    }
+    invalidateItem_ListWidget(d, d->dragItem);
+    d->dragItem = iInvalidPos;
+    return iTrue;
 }
 
 static iBool processEvent_ListWidget_(iListWidget *d, const SDL_Event *ev) {
@@ -333,10 +387,36 @@ static iBool processEvent_ListWidget_(iListWidget *d, const SDL_Event *ev) {
         d->noHoverWhileScrolling = iFalse;
     }
     if (ev->type == SDL_MOUSEMOTION) {
-        if (ev->motion.which != SDL_TOUCH_MOUSEID) {
-            d->noHoverWhileScrolling = iFalse;
+        const iInt2 mousePos = init_I2(ev->motion.x, ev->motion.y);
+        if (ev->motion.state == 0 /* not dragging */) {
+            if (ev->motion.which != SDL_TOUCH_MOUSEID) {
+                d->noHoverWhileScrolling = iFalse;
+            }
+            updateHover_ListWidget_(d, mousePos);
+        }
+        else if (d->dragItem != iInvalidPos) {
+            /* Start scrolling if near the ends. */
+            const int zone = 2 * d->itemHeight;
+            const iRect bounds = bounds_Widget(w);
+            float scrollSpeed = 0.0f;
+            if (mousePos.y > bottom_Rect(bounds) - zone) {
+                scrollSpeed = (mousePos.y - bottom_Rect(bounds) + zone) / (float) zone;
+            }
+            else if (mousePos.y < top_Rect(bounds) + zone) {
+                scrollSpeed = -(top_Rect(bounds) + zone - mousePos.y) / (float) zone;
+            }
+            scrollSpeed = iClamp(scrollSpeed, -1.0f, 1.0f);
+            if (iAbs(scrollSpeed) < 0.001f) {
+                stop_Anim(&d->scrollY.pos);
+                refresh_Widget(d);
+            }
+            else {
+                setFlags_Anim(&d->scrollY.pos, easeBoth_AnimFlag, iFalse);
+                setValueSpeed_Anim(&d->scrollY.pos, scrollSpeed < 0 ? 0 : scrollMax_ListWidget_(d),
+                                   scrollSpeed * scrollSpeed * gap_UI * 400);
+                refreshWhileScrolling_ListWidget_(d);
+            }
         }
-        updateHover_ListWidget_(d, init_I2(ev->motion.x, ev->motion.y));
     }
     if (ev->type == SDL_MOUSEWHEEL && isHover_Widget(w)) {
         int amount = -ev->wheel.y;
@@ -347,23 +427,45 @@ static iBool processEvent_ListWidget_(iListWidget *d, const SDL_Event *ev) {
         else {
             /* Traditional mouse wheel. */
             amount *= 3 * d->itemHeight;
-            moveSpan_SmoothScroll(&d->scrollY, amount, 200);            
+            moveSpan_SmoothScroll(
+                &d->scrollY, amount, 600 * scrollSpeedFactor_Prefs(prefs_App(), mouse_ScrollType));
         }
         return iTrue;
     }
     switch (processEvent_Click(&d->click, ev)) {
         case started_ClickResult:
             d->noHoverWhileScrolling = iFalse;
-            updateHover_ListWidget_(d, mouseCoord_Window(get_Window()));
+            updateHover_ListWidget_(d, mouseCoord_Window(get_Window(), ev->button.which));
             redrawHoverItem_ListWidget_(d);
             return iTrue;
         case aborted_ClickResult:
+//            endDrag_ListWidget_(d, pos_Click(&d->click));
+            if (d->dragItem != iInvalidPos) {
+                stop_Anim(&d->scrollY.pos);
+                invalidateItem_ListWidget(d, d->dragItem);
+                d->dragItem = iInvalidPos;
+            }
             redrawHoverItem_ListWidget_(d);
             break;
+        case drag_ClickResult:
+            if (d->dragItem == iInvalidPos && length_I2(delta_Click(&d->click)) > gap_UI) {
+                const size_t over = itemIndex_ListWidget(d, d->click.startPos);
+                if (over != iInvalidPos &&
+                    ((const iListItem *) item_ListWidget(d, over))->isDraggable) {
+                    d->dragItem = over;
+                    d->dragOrigin = sub_I2(topLeft_Rect(itemRect_ListWidget(d, over)),
+                                           d->click.startPos);
+                    invalidateItem_ListWidget(d, d->dragItem);
+                }
+            }
+            return d->dragItem != iInvalidPos;
         case finished_ClickResult:
+            if (endDrag_ListWidget_(d, pos_Click(&d->click))) {
+                return iTrue;
+            }
             redrawHoverItem_ListWidget_(d);
-            if (contains_Rect(innerBounds_Widget(w), pos_Click(&d->click)) &&
-                d->hoverItem != iInvalidSize) {
+            if (contains_Rect(itemRect_ListWidget(d, d->hoverItem), pos_Click(&d->click)) &&
+                d->hoverItem != iInvalidPos) {
                 postCommand_Widget(w, "list.clicked arg:%zu item:%p",
                                    d->hoverItem, constHoverItem_ListWidget(d));
             }
@@ -390,6 +492,7 @@ static void draw_ListWidget_(const iListWidget *d) {
     const int scrollY = pos_SmoothScroll(&d->scrollY);
     iPaint p;
     init_Paint(&p);
+    drawLayerEffects_Widget(w);
     drawBackground_Widget(w);
     alloc_VisBuf(d->visBuf, bounds.size, d->itemHeight);
     /* Update invalid regions/items. */ {
@@ -432,7 +535,9 @@ static void draw_ListWidget_(const iListWidget *d) {
                                                   init_I2(d->visBuf->texSize.x, d->itemHeight) };
                     beginTarget_Paint(&p, buf->texture);
                     fillRect_Paint(&p, itemRect, bg[i]);
-                    class_ListItem(item)->draw(item, &p, itemRect, d);
+                    if (index != d->dragItem) {
+                        class_ListItem(item)->draw(item, &p, itemRect, d);
+                    }
                     fillRect_Paint(&p, moved_Rect(sbBlankRect, init_I2(0, top_Rect(itemRect))), bg[i]);
                 }
             }
@@ -446,7 +551,9 @@ static void draw_ListWidget_(const iListWidget *d) {
                     const iRect      itemRect = { init_I2(0, j * d->itemHeight - buf->origin),
                                                   init_I2(d->visBuf->texSize.x, d->itemHeight) };
                     fillRect_Paint(&p, itemRect, bg[i]);
-                    class_ListItem(item)->draw(item, &p, itemRect, d);
+                    if (j != d->dragItem) {
+                        class_ListItem(item)->draw(item, &p, itemRect, d);
+                    }
                     fillRect_Paint(&p, moved_Rect(sbBlankRect, init_I2(0, top_Rect(itemRect))), bg[i]);
                 }
             }
@@ -457,6 +564,33 @@ static void draw_ListWidget_(const iListWidget *d) {
     }
     setClip_Paint(&p, bounds_Widget(w));
     draw_VisBuf(d->visBuf, addY_I2(topLeft_Rect(bounds), -scrollY), ySpan_Rect(bounds));
+    const iInt2 mousePos = mouseCoord_Window(get_Window(), 0);
+    if (d->dragItem != iInvalidPos && contains_Rect(bounds, mousePos)) {
+        iInt2 pos = add_I2(mousePos, d->dragOrigin);
+        const iListItem *item = constAt_PtrArray(&d->items, d->dragItem);
+        const iRect itemRect = { init_I2(left_Rect(bounds), pos.y),
+                                 init_I2(d->visBuf->texSize.x, d->itemHeight) };
+        SDL_SetRenderDrawBlendMode(renderer_Window(get_Window()), SDL_BLENDMODE_BLEND);
+        iBool dstOnto;
+        const size_t dstIndex = resolveDragDestination_ListWidget_(d, mousePos, &dstOnto);
+        if (dstIndex != d->dragItem) {
+            const iRect dstRect = itemRect_ListWidget(d, dstIndex);
+            p.alpha = 0xff;
+            if (dstOnto) {
+                drawRectThickness_Paint(&p, dstRect, gap_UI / 2, uiTextAction_ColorId);
+            }
+            else if (dstIndex != d->dragItem + 1) {
+                fillRect_Paint(&p, (iRect){ addY_I2(dstRect.pos, -gap_UI / 4),
+                                            init_I2(width_Rect(dstRect), gap_UI / 2) },
+                               uiTextAction_ColorId);
+            }
+        }
+        p.alpha = 0x80;
+        setOpacity_Text(0.5f);
+        class_ListItem(item)->draw(item, &p, itemRect, d);
+        setOpacity_Text(1.0f);
+        SDL_SetRenderDrawBlendMode(renderer_Window(get_Window()), SDL_BLENDMODE_NONE);
+    }
     unsetClip_Paint(&p);
     drawChildren_Widget(w);
 }
